import numpy as np
import torch
from torch.utils.data import Dataset


class BRCDataset(Dataset):
    """Dataset of sets."""

    def __init__(
        self,
        inputs: np.array,
        outputs: np.array,
    ) -> None:
        """Constructor.

        Args:
            inputs (np.array): Numpy array of input data.
            outputs (np.array): Numpy array of output data
        """
        self.input = inputs
        self.output = outputs

    def __len__(self) -> int:
        """Length of the data.

        Returns:
            int: Length of the data.
        """
        return len(self.input)

    def __getitem__(self, idx):
        """Generates a subset of data.

        Args:
            idx (tensor): Indices to be sampled.

        Returns:
            x (tensor): Tensor of input data.
            y (tensor): Tensor of corresponding outputs.
        """
        if torch.is_tensor(idx):
            idx = idx.tolist()

        x = torch.from_numpy(self.input[idx, :])
        y = torch.from_numpy(self.output[idx, :])

        return x, y



from typing import Tuple, List
import copy
import torch
import torch.nn as nn


class MultiLayerBase(nn.Module):
    """Executes a multilayer recurrent network given RNN cells for each layer."""

    def __init__(
        self,
        mode: str,
        cells: List[torch.nn.Module],
        hidden_size: int,
        batch_first: bool = True,
        bidirectional: bool = False,
        return_sequences: bool = False,
        device: torch.device = torch.
        device('cuda' if torch.cuda.is_available() else 'cpu'),
    ) -> None:
        """Constructor.

        Args:
            mode (string): Type of cell to be used in the recurrent network.
                Options are: ['GRU', 'LSTM', 'RNN', 'BRC', 'nBRC'].
            cell (List[torch.nn.Module]): List of cells to be used in each
            network layer. For example, torch.nn.GRUCell module initialised with
            the necessary arguments repeated num_layers times, i.e,
            [torch.nn.GRUCell(8,16),torch.nn.GRUCell(16,16)] for a 2 layered
            network with input_size = 8 and hidden_size = 16.
            NOTE: For bidirectional=True, remember to double the input_size for
            cells of layer 2 onwards, i.e, borrowing from the example above,
            [torch.nn.GRUCell(8,16),torch.nn.GRUCell(16*2,16)].
            hidden_size (int): Size of the hidden state.
            batch_first (bool): If True, then the input and output tensors are
                provided as (batch, seq_len, input_size). Default: False.
            bidirectional (bool): If True, becomes a bidirectional network, i.e,
                input is fed into 2 independent recurrent networks in the forward
                and backward direction. Default: False.
            return_sequences (bool): Return the hidden states for all
                time steps. Defaults to False.
            device (str): Whether to run model on GPU or CPU. Defaults to 'cuda',
            if available.
        """
        super().__init__()

        self.mode = mode

        self.layers = len(cells)
        self.hidden_size = hidden_size
        self.cells_forward = cells

        self.batch_first = batch_first
        self.return_seq = return_sequences
        self.device = device

        self.bidirectional = bidirectional

        self.num_directions = 2 if self.bidirectional else 1

        if self.bidirectional:
            self.cells_backward = copy.deepcopy(self.cells_forward)
            self.cells = nn.ModuleList(
                [
                    x for z in zip(self.cells_forward, self.cells_backward)
                    for x in z
                ]
            )
            assert len(self.cells) == self.layers * 2
        else:
            self.cells = nn.ModuleList(self.cells_forward)

    def init_state(self, batch_size: int) -> torch.Tensor:
        """Initialises the hidden/cell state for the recurrent network.

        Args:
            batch_size (int): Batch size of the state tensor initialised
                to 0s, should a new one be created.

        Returns:
            torch.Tensor: Tensor of 0s with shape
                (layers*num_directions, batch, hidden_size).
        """
        state = torch.zeros(
            (self.layers * self.num_directions, batch_size, self.hidden_size),
            dtype=torch.float32,
        ).to(self.device)
        return state

    def set_hidden_state(self, hidden_state: torch.FloatTensor) -> None:
        """Sets the hidden state for current cell.

        Args:
            hidden_state (torch.FloatTensor): Hidden state to be used
                in computation of the next hidden state.
        """

        self.hidden_state = hidden_state.to(self.device)

    def set_cell_state(self, cell_state: torch.FloatTensor) -> None:
        """Set the cell state for current cell.

        Args:
            cell_state (torch.FloatTensor): Cell state to be used
                in computation of the next cell and hidden states.
        """

        self.cell_state = cell_state.to(self.device)

    def multilayer_rnn(
        self,
        inputs: torch.Tensor,
        state: torch.Tensor = None,
    ) -> Tuple:
        """Multi-layer implementation of cells with only a hidden state.

        Args:
            inputs (torch.Tensor): Input tensor to be processed. If
                batch_first = True, (batch, seq_len, input_size) is
                expected, otherwise a tensor with shape
                (seq_len, batch, input_size) should be provided.
            NOTE: The operations in this function assume a 3-dimensional input.
                    For higher dimensions, please adapt the permutations
                    accordingly.
            state (torch.Tensor, optional): Tensor to intialise the hidden state
                with shape (layers*num_directions,batch,hidden_size).
                Default: None.
        Returns:
            Tuple: If return_sequence = True, returns an output tensor of shape
                (batch, seq_len, num_directions * hidden_size)
                for batch_first = True, and (seq_len, batch, num_directions * hidden_size)
                for batch_first = False, containing the output
                features h_t from the last layer of the network, for each t,
                as well as a tensor containing the hidden state for t = seq_len,
                with shape (layers * num_directions, batch, hidden_size).
                If return_sequence = False, returns the last hidden state from
                the last layer with shape (num_directions, batch, hidden_size).
        """
        if self.batch_first is False:
            inputs = inputs.permute(1, 0, 2)

        batch_size, sequence_length, input_size = inputs.size()

        last_hidden_states = []

        if state is None:
            hidden_state = self.init_state(batch_size)
        else:
            hidden_state = state

        for c in range(self.layers):
            layer_outputs = []

            for d in range(self.num_directions):

                stacked_states = []

                self.set_hidden_state(
                    hidden_state[self.num_directions * c + d]
                )

                range_length = (
                    range(sequence_length)
                    if d == 0 else reversed(range(sequence_length))
                )

                for i in range_length:

                    if c == 0:
                        inp = inputs[:, i, :]
                    else:
                        inp = output[:, i, :].to(self.device)

                    new_state = self.cells[self.num_directions * c +
                                           d](inp, self.hidden_state)

                    self.set_hidden_state(new_state)

                    stacked_states.append(new_state)

                layer_stacked = torch.stack(stacked_states).permute(1, 0, 2)

                last_hidden_states.append(layer_stacked[:, -1, :])

                if d == 1:
                    layer_stacked = layer_stacked.flip(1)

                layer_outputs.append(layer_stacked)

            output = torch.cat(layer_outputs, dim=-1)

        h_n = torch.stack(last_hidden_states)

        if self.batch_first is False:
            output = output.permute(1, 0, 2)

        if self.return_seq:
            return output, h_n

        else:
            return h_n[-2:, :, :
                       ] if self.bidirectional else h_n[-1, :, :].unsqueeze(0)

    def multilayer_lstm(self, inputs: torch.Tensor, states: Tuple) -> Tuple:
        """Multi-layer implementation of cells with a hidden and cell state.

        Args:
            inputs (torch.Tensor): Input tensor to be processed. If
                batch_first = True, (batch, seq_len, input_size) is
                expected, otherwise a tensor with shape
                (seq_len, batch, input_size) should be provided.
            NOTE: The operations in this function assume a 3-dimensional input.
                    For higher dimensions, please adapt the permutations
                    accordingly.
            state (Tuple, optional): Tuple of tensors to intialise the hidden
                and cell states with shape
                (layers * num_directions, batch, hidden_size). Default: None.
        Returns:
            Tuple: If return_sequence = True, returns an output tensor of shape
                (batch, seq_len, num_directions * hidden_size)
                for batch_first = True, and (seq_len, batch, num_directions * hidden_size)
                for batch_first = False, containing the output
                features h_t from the last layer of the network, for each t,
                as well as a tuple of tensors containing the hidden and cell
                states for t = seq_len, with each having a shape of
                (layers * num_directions, batch, hidden_size).
                If return_sequence = False, returns the hidden and cell state at
                t = seq_len from the last layer with shape
                (num_directions, batch, hidden_size).
        """

        if self.batch_first is False:
            inputs = inputs.permute(1, 0, 2)

        batch_size, sequence_length, input_size = inputs.size()

        last_hidden_states = []
        last_cell_states = []

        if states is None:
            hidden_state = self.init_state(batch_size)
            cell_state = self.init_state(batch_size)
        else:
            hidden_state, cell_state = states

        for c in range(self.layers):
            layer_hidden_outputs = []

            for d in range(self.num_directions):
                stacked_hidden_states = []
                stacked_cell_states = []

                self.set_hidden_state(
                    hidden_state[self.num_directions * c + d]
                )
                self.set_cell_state(cell_state[self.num_directions * c + d])

                range_length = (
                    range(sequence_length)
                    if d == 0 else reversed(range(sequence_length))
                )

                for i in range_length:

                    if c == 0:
                        inp = inputs[:, i, :]

                    else:
                        inp = output_hidden_states[:, i, :].to(self.device)

                    new_hidden_state, new_cell_state = self.cells[
                        self.num_directions * c +
                        d](inp, (self.hidden_state, self.cell_state))

                    self.set_hidden_state(new_hidden_state)
                    self.set_cell_state(new_cell_state)

                    stacked_hidden_states.append(new_hidden_state)
                    stacked_cell_states.append(new_cell_state)

                layer_hidden_stacked = torch.stack(stacked_hidden_states
                                                   ).permute(1, 0, 2)
                layer_cell_stacked = torch.stack(stacked_cell_states
                                                 ).permute(1, 0, 2)

                last_hidden_states.append(layer_hidden_stacked[:, -1, :])
                last_cell_states.append(layer_cell_stacked[:, -1, :])

                if d == 1:
                    layer_hidden_stacked = layer_hidden_stacked.flip(1)
                    layer_cell_stacked = layer_cell_stacked.flip(1)

                layer_hidden_outputs.append(layer_hidden_stacked)

            output_hidden_states = torch.cat(layer_hidden_outputs, dim=-1)

        h_n = torch.stack(last_hidden_states)
        c_n = torch.stack(last_cell_states)

        if self.batch_first is False:
            output_hidden_states = output_hidden_states.permute(1, 0, 2)

        if self.return_seq:
            return (
                output_hidden_states,
                (h_n, c_n),
            )
        else:

            return (
                (h_n[-2:, :, :], c_n[-2:, :, :]) if self.bidirectional else
                (h_n[-1, :, :].unsqueeze(0), c_n[-1, :, :].unsqueeze(0))
            )

    def forward(
        self, inputs: torch.Tensor, hidden_state: Tuple = None
    ) -> Tuple:
        """Executes the multilayer recurrent neural network based on the cell type.

        Args:
            inputs (torch.Tensor): Tensor to process.
            hidden_state (Tuple, optional): Tensor or tuple of tensor to initialise
                the recurrent cell states. Defaults to None.

        Returns:
            Tuple: The output returned by the network on processing the input.
            NOTE: In the bidrectional case, the hidden states at t=seq_len are
                simply the last returned hidden states in forward direction (t=n)
                in the reverse direction (t=0) for a given sequence with n steps.
                The output tensor is simply a concatenation of the hidden states
                at each time step t.
        """

        if self.mode in ['GRU', 'RNN', 'BRC', 'nBRC']:
            return self.multilayer_rnn(inputs, hidden_state)

        elif self.mode == 'LSTM':
            return self.multilayer_lstm(inputs, hidden_state)



import math
import torch
import torch.nn as nn
from typing import Tuple


class PeepholeLSTMCell(nn.Module):
    """LSTM with peephole connections."""

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        *args,
        **kwargs
    ) -> None:
        """Constructor.

        Args:
            input_size (int): Number of input features.
            hidden_sizes (dict): Number of hidden units in linear and
                encoder layers.
            bias (bool): Whether to include bias. Defaults to True.
        """
        super(PeepholeLSTMCell, self).__init__(*args, **kwargs)

        self.input_size = input_size
        self.hidden_size = hidden_size

        self.weights_x = nn.Parameter(
            torch.FloatTensor(self.input_size, self.hidden_size * 4)
        )
        self.weights_h = nn.Parameter(
            torch.FloatTensor(self.hidden_size, self.hidden_size * 4)
        )
        self.weights_c = nn.Parameter(
            torch.FloatTensor(self.hidden_size, self.hidden_size * 3)
        )

        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(self.hidden_size * 4))
        else:
            self.bias = torch.zeros((self.hidden_size * 4))

        self.init_params()

    def init_params(self) -> None:
        """Uniform Xavier initialisation of weights."""

        std_dev = math.sqrt(1 / self.hidden_size)
        for weight in self.parameters():
            weight.data.uniform_(-std_dev, std_dev)

    def init_hidden(self, batch_size: int) -> Tuple:
        """Initialise hidden and cell states.

        Args:
            batch_size (int) : batch size used for training.

        Returns:
            Tuple: a tuple containing the hidden state and cell state
                both initialized at 0s.
        """
        hidden_state = torch.FloatTensor(
            torch.zeros((batch_size, self.hidden_size))
        )
        cell_state = torch.FloatTensor(
            torch.zeros((batch_size, self.hidden_size))
        )

        return hidden_state, cell_state

    def forward(
        self, data_t: torch.Tensor, hidden_state: torch.Tensor,
        cell_state: torch.Tensor
    ) -> Tuple:
        """Single LSTM cell.

        Args:
            data_t (torch.Tensor): Element at step t with shape
                [batch_size, self.input_size]
            hidden_state (torch.Tensor): Hidden state of the LSTM cell of shape
                [batch_size, self.hidden_size]
            cell_state (torch.Tensor): Cell state of the LSTM cell of shape
                [batch_size, self.hidden_size]

        Returns:
            Tuple: A tuple containing the hidden state and cell state
                after the element is processed.
        """

        linear_xh = torch.matmul(data_t, self.weights_x) + torch.matmul(
            hidden_state, self.weights_h
        ) + self.bias

        linear_cxh = linear_xh[:, :self.hidden_size * 2] + torch.matmul(
            cell_state, self.weights_c[:, :self.hidden_size * 2]
        )

        forget_prob = torch.sigmoid(linear_cxh[:, :self.hidden_size])

        input_prob = torch.sigmoid(
            linear_cxh[:, self.hidden_size:self.hidden_size * 2]
        )

        candidates = torch.tanh(
            linear_xh[:, self.hidden_size * 2:self.hidden_size * 3]
        )

        cell_state = forget_prob * cell_state + input_prob * candidates

        linear_output = (
            linear_xh[:, self.hidden_size * 3:self.hidden_size * 4] +
            torch.matmul(
                cell_state,
                self.weights_c[:, self.hidden_size * 2:self.hidden_size * 3]
            )
        )

        output_gate = torch.sigmoid(linear_output)

        hidden_state = torch.tanh(cell_state) * output_gate

        return hidden_state, cell_state



import torch.nn as nn


class SelectItem(nn.Module):
    """Select output from a tuple to pass onto next function."""

    def __init__(
        self, item_index: int, elem_idx: int = None, batch_first: bool = True
    ) -> None:
        """Item selection module that can be included in nn.Sequential().

        Args:
            item_index (int): Index of the item to retrieve.
            elem_idx (int, optional): Index from which the selected item should
                be sliced. Depending on whether batch_first is True or False, the
                dimension along which it is sliced is either 0 or 1. Also,
                this implementation slices from the elem_idx until the end. E.g
                if elem_idx = -5, then the last 5 elements from the item are
                retained. Default: None.
            batch_first (bool, optional): Used as an indicator to slice along the
                correct dimensions using elem_idx.
        """
        super(SelectItem, self).__init__()
        self._name = 'selectitem'
        self.item_index = item_index
        self.elem_idx = elem_idx
        self.batch_first = batch_first

    def forward(self, inputs):
        """Selects item from tuple/list, slices it and returns it.

        Args:
            inputs (Tuple/List): Tuple/List from which item is to be retrieved.

        Returns:
            Tensor/Array: Selected and sliced (if elem_idx is given) Tensor/Array.
        """
        if self.elem_idx:
            if self.batch_first:
                return inputs[self.item_index][:, self.elem_idx:, :]
            else:
                return inputs[self.item_index][self.elem_idx:, :, :]
        else:
            return inputs[self.item_index]



import torch
import torch.nn as nn


class BistableRecurrentCell(nn.Module):
    """Bistable Recurrent Cell Layer (Vecoven et al. 2020)"""

    def __init__(self, input_size, output_dim, *args, **kwargs):
        """Constructor.

        Args:
            input_size (int): Number of input features.
            output_dim (int): Number of output features generated by the cell.
        """

        super(BistableRecurrentCell, self).__init__(*args, **kwargs)

        self.input_size = input_size
        self.output_dim = output_dim
        self.state_size = output_dim

        self.kernelz = nn.Parameter(
            torch.FloatTensor(self.input_size, self.output_dim)
        )
        self.kernelr = nn.Parameter(
            torch.FloatTensor(self.input_size, self.output_dim)
        )
        self.kernelh = nn.Parameter(
            torch.FloatTensor(self.input_size, self.output_dim)
        )

        self.memoryz = nn.Parameter(torch.FloatTensor(self.output_dim, ))
        self.memoryr = nn.Parameter(torch.FloatTensor(self.output_dim, ))

        self.bz = nn.Parameter(torch.FloatTensor(self.output_dim))
        self.br = nn.Parameter(torch.FloatTensor(self.output_dim))

        # Initialise the weights
        nn.init.xavier_uniform_(self.kernelz)
        nn.init.xavier_uniform_(self.kernelr)
        nn.init.xavier_uniform_(self.kernelh)

        nn.init.ones_(self.memoryz)
        nn.init.ones_(self.memoryr)

        nn.init.zeros_(self.bz)
        nn.init.zeros_(self.br)

    def forward(
        self, input_t: torch.Tensor, hidden_state: torch.Tensor = None
    ) -> torch.Tensor:
        """Bistable Recurrent Cell.

        Args:
            input_t (torch.Tensor): Input tensor at time step t.
            hidden_state (torch.Tensor): Hidden state at time time t-1.

        Returns:
            torch.Tensor: Tensor of the new hidden state at time t.
        """

        if hidden_state is None:
            hidden_state = self.hidden_state

        r = torch.tanh(
            torch.matmul(input_t, self.kernelr) + hidden_state * self.memoryr +
            self.br
        ) + 1
        z = torch.sigmoid(
            torch.matmul(input_t, self.kernelz) + hidden_state * self.memoryz +
            self.bz
        )
        output_hidden_state = z * hidden_state + (
            1.0 - z
        ) * torch.tanh(torch.matmul(input_t, self.kernelh) + r * hidden_state)

        return output_hidden_state

    def get_initial_state(self, batch_size=None, dtype=torch.float32):
        """Intialise hidden states.

        Args:
            batch_size (int, optional): Batch size of the input.
                Defaults to None.
            dtype (object, optional): Desired data type.
                Defaults to torch.float32.

        Returns:
            torch.Tensor: Hidden state intialised to 0.
        """
        return torch.zeros((batch_size, self.output_dim), dtype=dtype)



from typing import Tuple

import numpy as np
import torch
from torch.utils.data import Dataset


class BinaryAddition(Dataset):
    """Generates a dataset for binary addition."""

    def __init__(
        self, sample_size: int, sequence_length: int, max_sequence_length: int,
        min_sequence_length: int, mode: str
    ) -> None:
        """Constructor.

        Args:
            sample_size (int): Number of samples to generate.
            sequence_length (int): The sum of the lengths of the 2 binary
                numbers to add. Set it to 0 for mode = 'mixed'.
            max_sequence_length (int): Upper limit on length of the sequence.
            min_sequence_length (int): Lower limit length of the sequence.
            mode (string): 'single' generates a dataset for the given
                sequence length. 'mixed' generates a dataset containing all
                sequence lengths.
        """

        self.sample_size = sample_size
        self.max_seq_len = max_sequence_length
        self.min_seq_len = min_sequence_length
        self.sequence_length = sequence_length
        self.mode = mode
        self.x, self.y = self.create_dataset()

    def get_binary(self, length: int) -> str:
        """Generates a binary sequence.

        Args:
            length (int): Desired length of the binary sequence.

        Returns:
            string: binary sequence of specified length
        """
        return '{:0' + str(length) + 'b}'

    def create_dataset(self) -> Tuple:
        """Generates a dataset of binary numbers and their sums.

        Returns:
            Tuple: The input numbers and their corresponding sums are returned
                in binary integer form as a tuple.
        """

        if self.mode == 'single':
            # Input samples
            inputs = np.full(
                (self.sample_size, self.sequence_length + 1, 2), 0
            )
            # Target samples
            sums = np.full((self.sample_size, self.sequence_length + 1, 1), 0)

            for i in range(self.sample_size):

                max_len_nb1 = np.random.randint(1, self.sequence_length)
                max_len_nb2 = self.sequence_length - max_len_nb1

                # Generate random numbers to add
                nb1 = np.random.randint(2**(max_len_nb1 - 1), 2**(max_len_nb1))
                nb2 = np.random.randint(2**(max_len_nb2 - 1), 2**(max_len_nb2))
                sum_ = '{:0b}'.format(nb1 + nb2)

                inputs[i, :max_len_nb1, 0] = list(
                    reversed(
                        [
                            int(b)
                            for b in self.get_binary(max_len_nb1).format(nb1)
                        ]
                    )
                )
                inputs[i, :max_len_nb2, 1] = list(
                    reversed(
                        [
                            int(b)
                            for b in self.get_binary(max_len_nb2).format(nb2)
                        ]
                    )
                )
                sums[i, :len(sum_), 0] = list(reversed([int(b) for b in sum_]))

            return inputs, sums

        elif self.mode == 'mixed':
            sample_size = self.sample_size * (
                self.max_seq_len + 1 - self.min_seq_len
            )
            # Input samples
            inputs = np.full((sample_size, self.max_seq_len + 1, 2), 2)
            # Target samples
            sums = np.full((sample_size, self.max_seq_len + 1, 1), 2)

            for i in range(self.min_seq_len, self.max_seq_len + 1):
                for n in range(self.sample_size):
                    j = n + self.sample_size * (i - self.min_seq_len)
                    max_len_nb1 = np.random.randint(1, i)
                    max_len_nb2 = i - max_len_nb1

                    # Generate random numbers to add
                    nb1 = np.random.randint(
                        2**(max_len_nb1 - 1), 2**(max_len_nb1)
                    )
                    nb2 = np.random.randint(
                        2**(max_len_nb2 - 1), 2**(max_len_nb2)
                    )

                    sum_ = '{:0b}'.format(nb1 + nb2)

                    inputs[j, :max_len_nb1, 0] = list(
                        reversed(
                            [
                                int(b) for b in self.get_binary(max_len_nb1
                                                                ).format(nb1)
                            ]
                        )
                    )
                    inputs[j, :max_len_nb2, 1] = list(
                        reversed(
                            [
                                int(b) for b in self.get_binary(max_len_nb2
                                                                ).format(nb2)
                            ]
                        )
                    )
                    sums[j, :len(sum_), 0] = list(
                        reversed([int(b) for b in sum_])
                    )

            return inputs, sums

    def __len__(self) -> int:
        """Length of the data.

        Returns:
            int: Length of the data.
        """
        if self.mode == 'single':
            return self.sample_size
        elif self.mode == 'mixed':
            return len(self.x)

    def __getitem__(self, idx) -> Tuple:
        """Generates a subset of data.

        Args:
            idx (tensor): Indices to be sampled.

        Returns:
            x (tensor): Tensor of input binary numbers subset idx.
            y (tensor): Tensor of corresponding sums subset by idx.
        """

        if torch.is_tensor(idx):
            idx = idx.tolist()

        x = torch.from_numpy(self.x[idx, :, :]).float()
        y = torch.from_numpy(self.y[idx, :, :]).float()

        return x, y



"""Install package."""
import os
from setuptools import setup, find_packages

if os.path.exists('README.md'):
    long_description = open('README.md').read()
else:
    long_description = '''Pytorch Implementation of Bistable Recurrent Cell.'''

setup(
    name='brc_pytorch',
    version='0.1.3',
    license='MIT',
    description=('Pytorch Implementation of BRC.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/niklexical/brc_pytorch',
    author='Nikita Janakarajan, Jannis Born',
    author_email=('nikita.janakarajan907@gmail.com, jannis.born@gmx.de'),
    install_requires=['numpy', 'torch'],
    keywords=['PyTorch', 'Deep Learning', 'RNN', 'BRC'],
    packages=find_packages('.'),
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from .brc_layer import BistableRecurrentCell  # noqa
from .neuromodulated_brc_layer import NeuromodulatedBistableRecurrentCell  # noqa
from .peephole_lstm import PeepholeLSTMCell  # noqa
from .select_item import SelectItem  # noqa
from .multilayer_rnnbase import MultiLayerBase  # noqa